        page 250,150
;Comment--------------------------------------------------------
;
; Project name: PoFoCF
; --------------------
;
; Device driver for CompactFlash modules on an Atari Portfolio.
; version: v0.02
; date   : 20-01-2000
;
; Description:
; ------------
;
; This is the source code a device/interface driver. It is part
; of a hardware/software contraption that can be used to connect
; a CompactFlash module to a PortFolio. The hardware/software
; combination allows a PortFolio to access the Cf-module as a
; number of extra disk drives.
;
; Copyright notice:
; -----------------
;
; (C) 31-01-2000 by P.R. Faasse, Hakfort 635, 1102 LA Amsterdam
;
;   This program is free software; you can redistribute it and/or modify
;   it under the terms of the GNU General Public License as published by
;   the Free Software Foundation; either version 2 of the License, or
;   (at your option) any later version.
;
;   This program is distributed in the hope that it will be useful,
;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License
;   along with this program; if not, write to the Free Software
;   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; Warranty elaboration:
; ---------------------
; As nearly all computer programs this code comes  without any
; warranty. When using this code I am not responsible if anything
; goes wrong with your hardware/software, data, married life or
; otherwise.
;
; I've tried my very best to make this program work
;
; Bugs/error reporting:
; ---------------------
; Any reports of errors are welcome. I can again guarantee
; nothing, but as long as I'm interested in PortFolio's there is
; a very good chance that I'll do my best to kill any remaining
; bugs. I prefer error reports in e-mail form to:
;
;       faasse@nlr.nl
;
; If e-mail is not available to you you could also send a written
; report to:
;
;       P.R. Faasse
;       Hakfort 635
;       1102 LA Amsterdam
;       Netherlands
;
; History:
; --------
;
; 20-01-2000:
;       First version (0.01). The code has been copied mostly
;       from the pofoide v1.01 code. This thing should be far
;       easier to interface than a traditional harddisk, but the
;       harddisk driver code is quite ok, so i've copied the
;       existing code to build upon.
;
;       assembling/making the code:
;
;       MASM pofocf,,,,,
;       LINK pofocf,,,,
;       EXE2BIN pofocf.exe pofocf.sys
;
;
;       The complete command-line option set is:
;
;               /D              sets debug modus
;               /N              Skip FileSystem test
;
; 02-02-2000:
;       One quite marginal modification done: The filesystem tail
;       sizing code has been modified. This code is both simpler,
;       and more important: it makes a far better fit of the
;       CompactFlash module's capacity. The previous version of
;       PoFoCF was already quite bomb-free, i have good hope that
;       this version will do no less.
;
;End-------------------------------------------------------------

;Assembler------------------------------------------------------
;
; This is what MASM thinks is needed to start a simple device
; driver. I hardly understand most of it, seems it has something
; to do with segment-assignment.
;
code    segment para 'code'     ; define code segment
        assume cs:code, ds:code, es:code, ss:code
;
; nice and cryptic as you see, MASM is famous for that ;-)
;
;End------------------------------------------------------------

;Defs-----------------------------------------------------------
;
; The defines for the request packet interface
; They are really the offsets in the request packet of the
; parameter.
;
; request packet:
; ---------------
;
RqLen   equ     0       ; byte: length of the packet in bytes
RqUnit  equ     1       ; byte: unit number: 0 .. no of units
RqCmd   equ     2       ; byte: Command opcode byte
RqSts   equ     3       ; word: returned status
;
; 8 bytes reserved
;
RqHdr   equ     13      ; header hength in bytes
;
; start of the parameters:
; ------------------------
;
; Init call:
;
RqInitNUnits    equ     RqHdr+0 ; byte: Number of units
RqInitEndAdr    equ     RqHdr+1 ; ptr : End addres of driver
RqInitBPBarr    equ     RqHdr+5 ; ptr : BPB words array
RqInitCmdLine   equ     RqHdr+5 ; ptr : command line of driver
;
; Media check:
;
RqMChkDesc      equ     RqHdr+0 ; byte: Media descriptor
RqMChkRet       equ     RqHdr+1 ; byte return value
;
; BuildbPB:
;
RqBBPBDesc      equ     RqHdr+0 ; byte: Media descriptor
RqBBPBScr       equ     RqHdr+1 ; ptr : scratch area (512 bytes)
RqBBPBPtr       equ     RqHdr+5 ; ptr : BPB of the unit
;
; Read / Write (with/without verify and IoCtl....)
;
RqRdWrDesc      equ     RqHdr+0 ; byte: Media descriptor
RqRdWrIoAdr     equ     RqHdr+1 ; ptr : I/O address
RqRdWrCnt       equ     RqHdr+5 ; word: sector count
RqRdWrSec       equ     RqHdr+7 ; word: starting sector
;
; The error codes as returned to the DIP
;
EWriteProt      equ     0       ; write-protect error
EUNIT           equ     1       ; Unknown unit error
ENRDY           equ     2       ; drive not ready error
ECMD            equ     3       ;
ECRC            equ     4       ;
ELEN            equ     5       ;
ESEEK           equ     6       ;
EMEDIA          equ     7       ;
ESECNF          equ     8       ;
ENoPaper        equ     9       ; ok, the CF has no paper
                                ;     problems...
EWRITE          equ     10      ;
EREAD           equ     11      ;
EGENERAL        equ     12      ;
ERes1           equ     13      ;
Eres2           equ     14      ;
EDISKCH         equ     15      ;
;
;End------------------------------------------------------------

;ORG------------------------------------------------------------
        ORG 0           ; device drivers have no PSP
;End------------------------------------------------------------

;Dev Header-----------------------------------------------------
;
; This header is interpreted by DOS/DIP to find out about the
; device driver.
;
;---------------------------------------------------------------
;
            dd 0FFFFFFFFh               ; Pointer to next driver
            dw 6000h                    ; Attribute word
            dw Offset StratRoutine      ; ptr to strategy routine
IntVektor   dw Offset IntRoutine        ; ptr to interrupt routine
Units       db 23                       ; number of units (max)
;
            db 'PoFoCF '        ; OEM (that's us here..) string
;
;
;End------------------------------------------------------------

;---------------------------------------------------------------
; Strategy-Routine:
;
; Does nearly nothing, it just saves the pointer to the request
; block in a local parameter
;
StratRoutine:
            mov cs:word ptr DosPointer,bx       ; save pointer to
            mov cs:word ptr DosPointer+2,es     ; request block
            retf                                ; exit
;
;---------------------------------------------------------------
; Dummy routine:
;---------------------------------------------------------------
;
; The dummy routine does nothing. It is needed because the PoFo
; would crash if it was not there. It simply returns and error,
; Unfortunately it is not possible to skip the complete
; installation of the driver on a PoFo.
;
; NB: I'm not so sure about that anymore.. making BuildBPB return a wrong
; sector size could just do that?...
;
DummyRoutine:
        push    di                      ;
        push    es                      ;
        les     di,cs:[DosPointer]      ; es:di -> request block
        mov     es:[di+RqSts],8102h     ; drive not ready, done bit
        pop     es                      ;
        pop     di                      ;
        retf                            ;
;
;End------------------------------------------------------------

;Static---------------------------------------------------------
;
; the static storage of the device driver.
;
;---------------------------------------------------------------
;
Debug           db      0       ; debug flag <>0 -> debug messages
NoFsChk         db      0       ; skip fs check (1=skip)
DoIdent         db      1       ; Flag for Do Disk Ident 1=do it
DoInit          db      1       ; Flag for Init 1=do init
errorc          db      0       ; error flag/byte (<>0 -> error)
DosPointer      dd      0       ; request block pointer
DataPointer     dd      0       ; data I/O addres pointer
;
; the array of BPB pointers as specified in the:
;
;       DOS4.0 programmers manual
;
BPBarray:
        dw      offset BPB ; D: ; For the moment I point them all
        dw      offset BPB ; E: ; at my one-and-only BPB.
        dw      offset BPB ; F: ; I have made room for all
        dw      offset BPB ; G: ; the remaining alfabet.
        dw      offset BPB ; H: ; I'll manipulate the dsize
        dw      offset BPB ; I: ; parameter of my BPB to fit
        dw      offset BPB ; J: ; the real BPB parameter(s) of
        dw      offset BPB ; K: ; the current unit.
        dw      offset BPB ; L: ;
        dw      offset BPB ; M: ;
        dw      offset BPB ; N: ; <- this is where my 351 MBytes
        dw      offset BPB ; O: ;    disk ends
        dw      offset BPB ; P: ;
        dw      offset BPB ; Q: ;
        dw      offset BPB ; R: ;
        dw      offset BPB ; S: ;
        dw      offset BPB ; T: ;
        dw      offset BPB ; U: ;
        dw      offset BPB ; V: ;
        dw      offset BPB ; W: ;
        dw      offset BPB ; X: ;
        dw      offset BPB ; Y: ;
        dw      offset BPB ; Z: ; room for 23 units
;
; I decided to make a LockRdWr bit per unit after all. It's a
; little more work, but the rigid solution I had before would
; make even me wild...
;
LockBits:       dd      0       ; 32 lock bits.
;
;---------------------------------------------------------------
; The BPB (Bios Parameter Block). This block of data is passed
; (pointed to) when the DIP wants to know the disk's geometry.
;---------------------------------------------------------------
;
; The parameters of the file-system, I support 12-bits FATs for
; the moment. So does DIP, another DOS may be better at this
; trick. NB: the parameter dsize in the real BPB buffer is
; updated each time a SetBuffer is requested.
;
rootdir equ     128             ; # entries in the root directory
fats    equ     1               ; # fat's
cluster equ     16              ; sectors/cluster
dtype   equ     0F8H            ; media type byte (F8H = harddisk)
dsize   equ     65280           ; number of sectors in the disk
;
; the fixed parameters of the DOS file-system
;
secsize equ     0200H           ; always 512 bytes/sector
dirsize equ     020H            ; bytes per directory entry
dskresv equ     1               ; always one reserved sector (why?)
;
; the pars I compute from the others
;
; The size of the FAT in sectors
;
; fatsize equ ((3 * dsize)+1)/(2 * secsize * cluster) ; sectors/fat
;
fatsize         equ     12      ; MASM is no good calculator
;
; The size of the root directory in sectors
;
rootsize        equ     (rootdir * dirsize) / secsize
;
; admin size of a file-system:
; FAT, rootdir reserved and bootblock (the '1')
;
admsize         equ     (fatsize + rootsize + dskresv + 1)
;
; When the disk tail size is below a certain lower limit I do not
; bother to start a new partition. A partition with ONLY a
; bootblock and a FAT seems like nonsense to me.
;
minsize equ     40             ; minimum sensible partition
                               ; for pofocf: 20 KBytes
;
; I fill in the BPB for the disk that is being accessed. That
; happens when a read/write or a BuildBPB is requested.
;
; The Normal and Last disk size
;
Cunit           db      0               ; current unit being used
DSizeN          dw      dsize           ; Normal disk size
DSizeL          dw      0               ; last-unit disk size
ClustN          db      cluster         ; Normal cluster size
ClustL          db      0               ; last-unit cluster size
FatSizeN        dw      fatsize         ; Normal FAT size
FatSizeL        dw      0               ; last-unit FAT size
;
; I check this with data from disk. For simplicity I compare
; the entire head of the boot-block to here. That saves some
; counting.
;
BootBlock:
        jmp     short noboot    ; short jmp ends up here
        db      0               ; filler byte
;
; The BPB itself is where the parameters end up. NB: The ones
; marked (fixed: <value>) are checked when a TstFs is executed.
; This is done when the driver Init-s. If one of these parameters
; is found different from the <value> the entire disk is read/write
; locked. This can be made ok by executing a mkfs on the
; offending partition(s). A subsequent reboot will re-check the
; parameters and remove the read/write lock when all partitions
; are found to have a valid file-system. This is perhaps a bit
; too rigid, but it does assure that there are no reads/writes to
; partitions that do not contain a valid file-system.
;
BPBNam: db      "PoFoCF  " ;    ; volume name (8 chars)
;                           offs in BPB:
BPB     dw      secsize    ;+00 ; sector size     (fixed: 0200H)
BPBclus db      cluster    ;+02 ; cluster size    (fixed: 16)
BPBResv dw      dskresv    ;+03 ; reserved sectors (mostly 1)
BPBfatn db      fats       ;+05 ; no of FAT's     (fixed: 1)
BPBroot dw      rootdir    ;+06 ; root dir entries
BPBdsiz dw      dsize      ;+08 ; total number of sectors
BPBdtyp db      dtype      ;+0A ; media code      (fixed: F8H = hard-disk)
BPBfats dw      fatsize    ;+0B ; sectors per FAT (fixed: 12)
        dw      8          ;+0D ; sectors per track (nonsense
        dw      2          ;+0F ; number of heads     really)
        dd      0          ;+11 ; hidden sectors

        ; some reserved space, 11 bytes in all. This should
        ; satisfy even DOS4.0 and on
        db      0,0,0,0,0       ; reserved
        db      0,0,0,0,0,0     ; reserved too
;
noboot:
;
bootsize        equ     (offset noboot - offset bootblock)
;
;---------------------------------------------------------------
;
; For debugging, I've added a name table mechanism to the diver.
; This part will not be in the real code when a production
; version goes into the world. It saves a hell off a lot of
; trouble when debugging. My nice little debugger (ddt) will find
; this and use it to make me find out where I'm looking.
;

        ; name table entry definition macro
NM      macro   NtName
        dw      NtName
        db      '&NtName',0
        endm

; un-comment the next line to get a name table. With this
; line commented you'll not get a name table:
;With_NT equ    1        ; names tabel ON

; check if the names table is requested.
IFDEF With_NT

        ; signal string for ddt. It will recognize the
        ; existence of the table when it finds this string.
        db      'ddt nt'        ; for ddt: name table follows

        ; The macro can make life a load easier here
        ; simply: NM <name of label to put in the name table>
        ; Note: comment is allowed here...
        NM      ReadSec         ; read one sector
        nm      WriteSec        ; write a sector
        nm      IntRoutine      ; interrupt routin
        nm      Read            ; read entry point
        nm      SetBuffer       ; setup buffers
        nm      RdWrError       ; error processing
        nm      ReadMore        ; more to read
        nm      RdWrDone        ; read done
        nm      WaitNBsy        ; Wait till disk is not busy
        nm      WaitDRDY        ; same thing
        nm      FixBPB          ; make BPB ok
        nm      SetLBB          ; unit -> LBB conversion
        nm      ReadSecErr      ; handles read-sector errors
        nm      SetLBA          ; set LBA in the disk
        nm      DoCmd           ; write command to disk
        nm      ReadSts         ; read disk's status register
        nm      ReadSecStsErr   ; handles status errors
        nm      WaitDRQ         ; Wait till disk requests data
        nm      ReadSecDrqErr   ; error processing thereof
        nm      ReadBlock       ; read a data block from the disk
        nm      WriteBlock      ; write a block to disk
        nm      ReadByte        ; read byte from CF
        nm      WriteByte       ; write a byte to CF
        nm      ReadReg         ; read a register
        nm      ReadWord        ; read a word from CF data
        nm      RdSkip          ; read/skip data
        nm      SoftReset       ; execute soft-reset
        nm      WaitTimOut      ; wait time-out
        nm      FinWait         ; wait finished

        ; Init routines too
        nm      ParsePars       ; Parse run-string pars
        nm      InitCf          ; init CF module
        nm      DisChar         ; display a character
        nm      ToHex           ; convert to hex
        nm      DisHexByte      ; display hex byte
        nm      DisHexWord      ; display a hex word
        nm      NewLine         ; goto new line
        nm      TestFs          ; test file-system
        nm      CheckFailedStr  ; check failed
        nm      CheckIO         ; test if PoFo & drive
        nm      Number          ; read a number
        nm      Parameter       ; parse parameters
        nm      ParamError      ; error in parameter

        ; Init parameters
        nm      TestBlock       ; I/O testblock
        nm      Tst             ; I/O test block too

        ; Test: the parameters too, the new ddt should print them
        ; too
        nm      BPB             ; BPB
        nm      Debug           ; debug flag <>0 -> debug messages
        nm      DoIdent         ; Flag for Do Disk Ident 1=do it
        nm      NoFsChk         ; skip fs check (1=skip)
        nm      DoInit          ; Flag for Init 1=do init
        nm      errorc          ; error flag/byte (<>0 -> error)
        nm      DosPointer      ; request block pointer
        nm      DataPointer     ; data I/O addres pointer

        ; CF interface parameters too
        nm      LBB             ; LBA base (Partition start)
        nm      LBA             ; Lineair Block Address (LBA)
        nm      SCnt            ; sector count.
        nm      CUnit           ; Current Unit number

        ; end of the name table
        db      0,0,0           ; end of the names table

ENDIF
;
; When the initialisation of the driver goes ok the code after
; the label EndResident will be released from memory. If an error
; in the initial phase of the driver happens the part after
; ErrEndResident will be released. This saves some of the very
; valuable memory space of the Pofo. In this error case ONLY the
; the device header, strategy routine, the DummyRoutine and the
; BPB array and BPB itself  will remain in the memory.
;
ErrEndResident:
;
;End-------------------------------------------------------------

;Despatcher------------------------------------------------------
;
; This is the 'interrupt' part of the driver. It parses the
; command and executes the appropriate service routine. It is NOT
; an interrupt-driven routine in the sense that hardware-
; interrupts activate this code. Dos/Dip have all the provisions
; for multi-tasking device drivers, just the real multi-tasking
; kernel is not there....
;
IntRoutine:

        push    ax              ; save all registers
        push    bx              ;
        push    cx              ;
        push    dx              ;
        push    si              ;
        push    di              ;
        push    bp              ;
        push    ds              ;
        push    es              ;
        pushf                   ;

        mov     ax,cs           ;
        mov     ds,ax           ; ds:=cs

        ; prepare pointer to request packet
        les     di,[DosPointer] ; make es:di -> request packet

        ; clear error code
        mov     errorc,0        ; preset error to 0

        ; parse the command of the request packet
        mov     bl,es:[di+RqCmd]; get command byte
        xor     bh,bh           ; BX = Command code

        cmp     bl,12           ; test if acceptable command
        jbe     Cmd_ok          ; if so, continue

        ; this driver supports commands 0 .. 12
        mov     errorc,ECMD     ; unknown Command
        jmp     CmdRet          ; exit with error code set
;
; The main command interpreter of the driver. It uses the
; JumpTable to find which routine to call. One rouine for each
; supported function call. Upon being called the subroutine
; pointed to by bx the following registers have the following
; contents:
;
;  ds    = local code/data segment
;  es:di -> request packet
;
; The handler can signal errors by setting a non-zero error code
; in the parameter errorc. The value here is the error code that
; will be passed back to DOS/DIP. See the DOS/DIP manual for a
; description of error codes.
;
Cmd_ok: shl     bx,1                    ; it is a word table
        call    word ptr [JumpTable+bx] ; call function handler

        ; entry when the command has done it's work
CmdRet:

        ; make sure you NEVER return an error for an INIT
        ; call. The PoFo crashes like a bomb when that happens
        cmp     byte ptr es:[di+RqCmd],0        ;
        jnz     CmDoErr                         ;

        ; an INIT call NEVER returns an error
        mov     errorc,0        ; whatever happens...

        ; process the value of errorc
CmDoErr:mov     al,errorc       ; test if no errors
        mov     ah,0            ;
        or      al,al           ;
        jz      CmNoErr         ;

        mov     ah,80h          ; set error flag
CmNoErr:or      ah,01           ; set done flag
        mov     es:[di+RqSts],ax; set the return status

        popf                    ; restore all registers
        pop     es              ;
        pop     ds              ;
        pop     bp              ;
        pop     di              ;
        pop     si              ;
        pop     dx              ;
        pop     cx              ;
        pop     bx              ;
        pop     ax              ;

        retf                    ; far return
;
; The JumpTable of the main command interpreter.
;
JumpTable:

        DW      Offset Init             ; Command  0
        DW      Offset MediaChk         ;          1
        DW      Offset BuildBPB         ;          2
        DW      Offset IoctlRead        ;          3
        DW      Offset Read             ;          4
        DW      Offset Dummy            ; unsup    5
        DW      Offset Dummy            ; unsup    6
        DW      Offset Dummy            ; unsup    7
        DW      Offset Write            ;          8
        DW      Offset WriteVfy         ;          9
        DW      Offset Dummy            ; unsup   10
        DW      Offset Dummy            ; unsup   11
        DW      Offset IoctlWrite       ;         12

;End----------------------------------------------------------

;Handlers-----------------------------------------------------
;
; The function handlers of the interrupt routine
;
;-------------------------------------------------------------

;-------------------------------------------------------------
; Handler Dummy
; purp: dead end for all non-supported function calls
; in  : nothing
; out : error code set to indicate that the command is not
;       supported
; uses: nothing

Dummy:  mov     errorc,ECMD     ; unknown command error
        ret                     ;

;-------------------------------------------------------------
; Handler Init
; purp: Initialise the driver.
; in  : nothing
; out : driver initialised and request block parameters filled
;       in. NB: when something goes wrong in this initial phase
;       the number of devices is set to 0 and the Dummy routine
;       is hooked up as the interrupt routine. This is
;       self-modifying code, the entry for IntRoutine in the
;       driver's header is modified.
; uses: registers, request header etc...
;
; NB: This handler could be placed in the transient part of the
;     driver. It would not take resident code space then. It has
;     been kept here because all the other handlers are here too.
;     All it's routines are in the transient part of the driver
;     (at the tail of this code..).

Init:
        ; issue start message
        mov     dx,Offset InitStr       ;
        mov     ah,9                    ;
        int     21h                     ;

        ; test if init is to be done (skip most of the work
        ; if it is called a seccond time...)
        cmp     DoInit,1                ;
        jnz     InitOk                  ;

        ; parse run-string parameters
        call    ParsePars               ; (see tail of the code)
        jc      InitErr                 ;

        ; init the CF bus, check if there is any drive
        ; connected.
        call    CheckIO                 ; (see tail of the code)
        jc      InitErr                 ;

        ; test if any units found
        cmp     Units,0                 ;
        jz      InitErr                 ;

        ; all CF start things done, setup return parameters
        jmp     InitOk                  ;

        ; In case of error, ALL function calls end up at the dummy
        ; routine; This will result in the PoFo reporting an error
        ; without a total system crash. Unfortunately the PoFo WILL
        ; install the driver,
        ; This is NOT in accordance with the normal DOS
        ; behaviour, but we'll have to live with that (or burn a
        ; modified PoFo ROM....).

InitErr:mov     units,1                         ;
        mov     IntVektor,offset DummyRoutine   ;

        ; in case of errors the driver's resident part will
        ; become noticably smaller. All code AFTER the dummy
        ; routine will be removed from the memory.
        mov     ax,offset ErrEndResident        ;
        jmp     InitTail                        ;

InitOk: mov     ax,offset EndResident           ;
;       jmp     InitTail                        ;

        ; finish up things in the driver's return packet from the
        ; INIT call; ax = end of resident portion

InitTail:

        mov     word ptr es:[di+RqInitEndAdr],ax        ; setup mem pointer
        mov     word ptr es:[di+RqInitEndAdr+2],cs      ; to driver's end

        ; setup BPB pointer
        mov     word ptr es:[di+RqInitBPBarr],offset BPBarray
        mov     word ptr es:[di+RqInitBPBarr+2],cs

        ; return number of units
        mov     al,Units                        ;
        mov     byte ptr es:[di+RqInitNUnits],al;

        ; make sure you do the real Init work only once,
        ; a seccond time the code would no longer be there...
        mov     DoInit,0                        ;

        ret                                     ;

;-------------------------------------------------------------
; Handler MediaChk
; purp: test if device medium changed.
; in  : nothing
; out : do not know if medium changed, for all units.
; uses: nothing

MediaChk:
        mov     byte ptr es:[di+RqMChkdesc],dtype       ;
        mov     byte ptr es:[di+RqMChkRet],0H           ;
        ret                                             ;

;-------------------------------------------------------------
; Handler BuildBPB
; purp: get drive parameters (DOS/DIP-level)
; in  : nothing
; out : pointer to BPB set in return parameters
; uses: nothing
; NB  : this routine also does some strange manipulations
;       with the BPB's sector size. Seems to have something to
;       do with DIP's problems with >512 bytes clusters.

BuildBPB:

        ; prepare returned buffer
        mov     byte ptr es:[di+RqBBPBDesc],dtype           ;
        mov     word ptr es:[di+RqBBPBPtr],Offset BPB       ; pointer to BPB
        mov     word ptr es:[di+RqBBPBPtr+2],cs             ;

        ; set the correct parameters for this unit in the BPB
        mov     al,byte ptr es:[di+RqUnit]      ; get unit number
        call    FixBPB                          ; make BPB ok

        ret                             ;

;-------------------------------------------------------------
; Handler IoCtlRead
; purp: read parameters from the driver, driver present signal
; in  : nothing
; out : driver's parameters in the return block.
; uses: some registers
;

IoctlRead:

        ; setup data
        mov     al,byte ptr es:[di+RqUnit]      ; get unit number
        mov     CUnit,al                ;
        call    SetLBB                  ; make LBB ok
        call    FixBPB                  ; make BPB ok

        ; get pointer to the return packet data block
        cld                             ;
        push    es                      ;
        push    di                      ;
        les     di,es:[di+RqRdWrIoAdr]  ;

        ; fill in the return packet data
        mov     ax,'KP'                 ; signal bytes
        stosw                           ;
        mov     al,CUnit                ; unit number
        stosb                           ;

        ; I pass a load of internal information back too. This is
        ; quite nice for on-line disgnostics of the driver.

        ; the partition's base
        mov     ax,word ptr LBB         ;
        stosw                           ;
        mov     ax,word ptr LBB+2       ;
        stosw                           ;

        ;  the offset of the partition's bootblock info section
        mov     ax,cs                   ; the driver's code-segment
        stosw                           ;
        mov     ax,offset BootBlock     ;
        stosw                           ;

        ; the disk size
        mov     ax,word ptr SPD         ; disk's SPD
        stosw                           ;
        mov     ax,word ptr SPD+2       ;
        stosw                           ;

        mov     ax,0                    ; dummy CHS config
        stosw                           ;
        stosw                           ;
        stosw                           ;

        ; information about the latest access to the disk
        mov     ax,word ptr LBA         ; in LBA form
        stosw                           ;
        mov     ax,word ptr LBA+2       ;
        stosw                           ;
        mov     ax,0                    ; dummy CHS access
        stosw                           ;
        stosw                           ;
        stosw                           ;

        ; scan the IDE I/O ports
        mov     ax,0                    ; give dummy results
        stosw                           ; keep old ddiag happy
        stosb                           ;

        ; read all relevant IDE registers. Note that this will
        ; destroy the previously read I/O registers values
        mov     ax,CFLBA1               ;
        call    ReadReg                 ;
        stosb                           ;
        mov     ax,CFLBA2               ;
        call    ReadReg                 ;
        stosb                           ;
        mov     ax,CFLBA3               ;
        call    ReadReg                 ;
        stosb                           ;
        mov     ax,CFLBA0               ;
        call    ReadReg                 ;
        stosb                           ;
        mov     ax,CFNum                ;
        call    ReadReg                 ;
        stosb                           ;
        mov     ax,CFSts                ; status register
        call    ReadReg                 ;
        stosb                           ;
        mov     ax,CFErr                ;
        call    ReadReg                 ;
        stosb                           ;

        ; done, get registers ok again
        pop     di                      ;
        pop     es                      ;
        ret                             ;

;-------------------------------------------------------------
; Handler Read
; purp: read sectors from the disk
; in  : es:di -> request packet
; out : data written into the indicated buffer, error flag
; uses: registers

Read:

DoRead: call    SetBuffer               ; get parameters from
                                        ; request block
ReadMore:

        call    ReadSec                 ; read one sector
        jc      RdWrError               ; handle errors

        add     word ptr LBA,1          ; next sector
        adc     word ptr LBA+2,0        ;
        dec     SCnt                    ; test if any more to
        cmp     SCnt,0                  ; read
        jnz     ReadMore                ;
        jmp     RdWrDone                ;

;-------------------------------------------------------------
; Handler IoCtlWrite
; purp: write a single sector to the disk
; in  : parameters in the request block
; out : error code
; uses: registers

IoctlWrite:

        ; sector write, always one sector, no lock check

        ; make ds:bx -> data buffer
        lds     bx,es:[di+RqRdWrIoAdr]  ;

        mov     ax,[bx+513]     ; get sector number
        xchg    ah,al           ; swap to get the
                                ; byte-order ok

        mov     word ptr es:[di+RqRdWrSec],ax   ; put in the request header
        mov     word ptr es:[di+RqRdWrCnt],1    ;

        mov     ax,cs                   ; to fake a normal write
        mov     ds,ax                   ; ds = cs again
        jmp     DoWrite                 ; go write the sector,
                                        ; ignore lockout mech.

;-------------------------------------------------------------
; Handler Write/WriteVfy
; purp: write sectors to the disk
; in  : es:di -> request packet
; out : data written to the disk, error flag
; uses: registers

WriteVfy:   ; This write-command is used by DIP/DOS when verify has
            ; been set to ON. I just write. Verify could be
            ; implemented later on.

Write:

        ; test lockout mechanism
        mov     al,es:[di+RqUnit]       ; unit code
        call    GetLock                 ; get lock bit
        jc      RdWrLocked              ;

        ; go do writes
DoWrite:call    SetBuffer               ; get parameters

WriteMore:

        call    WriteSec                ; write one sector
        jc      RdWrError               ; handle errors

        add     word ptr LBA,1          ; next sector
        adc     word ptr LBA+2,0        ;
        dec     SCnt                    ; test if any more to
        cmp     SCnt,0                  ; write
        jnz     writemore               ;
        jmp     RdWrDone                ;

RdWrLocked:

        ; read/write lockout, give error & exit
        mov     errorc,ENRDY            ; drive not ready
        mov     ax,es:[di+RqRdWrCnt]    ; nothing done
        mov     SCnt,ax                 ;

RdWrError:

        ; error handler for read/write. Error code
        ; is already set. I just fill in what has been
        ; read or written.
        mov     ax,es:[di+RqRdWrCnt]    ; get requested count
        sub     ax,SCnt                 ; what has been Rd/Wr
        mov     es:[di+RqRdWrCnt],ax    ;

RdWrDone:

        ; make the I/O ports passive
        ret                             ; done

;
;End------------------------------------------------------------

;Routines-------------------------------------------------------
;
; The subroutines of the handlers:
;
;---------------------------------------------------------------

;----------------------------------------------------------------
; Routine SetBuffer
; purp: copies incoming parameters from request block to local
;       parameters.
; in  : data in request block from DIP:
;       ds = cs
;       ds:DosPointer -> request packet.
; out : data in local parameters:
;       LBB = base-sector number of the unit.
;       LBA = sector number of first sector to transfer
;       SCnt = number of sectors to transfer
;       BPB made ok for this unit
;       DataPointer -> I/O data address
;
; uses: ax
;

SetBuffer:

        ; get the number of sectors
        mov     ax,es:[di+RqRdWrCnt]    ; get number of sectors
        mov     SCnt,ax                 ;

        ; convert unit code to LBB
        mov     al,es:[di+RqUnit]       ; unit code
        mov     CUnit,al                ; store
        call    SetLBB                  ; convert to LBB

        ; load the BPB with the dsize for the current unit
        mov     al,es:[di+RqUnit]       ; RqUnit [0 .. units-1]
        call    FixBPB                  ; make BPB ok

        ; now, make the LBA
        mov     ax,es:[di+RqRdWrSec]    ; get starting sector number
        add     ax,word ptr LBB         ; + LBB -> LBA
        mov     word ptr LBA,ax         ;
        mov     ax,word ptr LBB+2       ;
        adc     ax,0                    ;
        mov     word ptr LBA+2,ax       ;

        ; put the data pointer in local buffer
        mov     ax,word ptr es:[di+RqRdWrIoAdr]         ;
        mov     word ptr DataPointer,ax                 ;
        mov     ax,word ptr es:[di+RqRdWrIoAdr+2]       ;
        mov     word ptr DataPointer+2,ax               ;

        ret                             ;

;----------------------------------------------------------------
; Routine FixBpb
; purp: fix up the BPB for a specified drive
; in  : al = unit number
; out : BPB made ok for that unit. Delivers the standard
;       parameters when non-last unit requested, special ones for
;       the last unit of the disk.
; uses: ax

FixBPB:

        ; save input pars
        push    ax                      ;

        ; al = unit no 0 .. units-1
        inc     al                      ; make it [1 .. units]
        cmp     al,units                ;
        jz      fixlastunit             ;

        ; not the last unit, make BPB for normal unit
        mov     ax,DSizeN               ; standard disk size
        mov     BPBdsiz,ax              ;
        mov     al,ClustN               ; standard cluster size
        mov     BPBclus,al              ;
        mov     ax,FatSizeN             ; standard FAT size
        mov     BPBfats,ax              ; 
        jmp     FixBPBEnd               ;

        ; prepare the last unit's BPB
fixlastunit:
        mov     ax,DSizeL               ; disk-size for Last
        mov     BPBdsiz,ax              ;
        mov     al,ClustL               ; cluster size for Last
        mov     BPBclus,al              ;
        mov     ax,FatSizeL             ; fat size for last
        mov     BPBfats,ax              ;
;       jmp     FixBPBEnd               ;

FixBPBEnd:

        ; restore registers
        pop     ax                      ;
        ret                             ;

;----------------------------------------------------------------
; Routine SetLBB
; purp: convert unit number to LBB value
; in  : al = unit number
; out : LBB set for that unit
; uses: nothing

SetLBB: push    ax                      ; save input
        push    cx                      ;
        push    dx                      ;

        mov     ah,0                    ;
        mov     cx,dsize                ;
        mul     cx                      ; x disk size
        mov     word ptr LBB,ax         ; put in buffer
        mov     word ptr LBB+2,dx       ; for LBB

        pop     dx                      ;
        pop     cx                      ;
        pop     ax                      ;
        ret                             ;

;----------------------------------------------------------------
; Routine SetLock
; purp: Set a read/write lock bit for a unit
; in  : al = unit number
; out : read/write lock bit set for that unit
; uses: flags

SetLock:push    ax                      ; save input
        push    cx                      ;

        ; test high/low word
        cmp     al,16                   ;
        jnc     setuplk                 ; upper byte lock bit

        ; lock bits 0 .. 15 in lower word
        mov     cl,al                   ;
        mov     ax,1                    ;
        shl     ax,cl                   ;
        or      word ptr LockBits,ax    ;
        jmp     setlken                 ;

setuplk:sub     al,16                   ;
        mov     cl,al                   ;
        mov     ax,1                    ;
        shl     ax,cl                   ;
        or      word ptr LockBits+2,ax  ;

setlken:pop     cx                      ;
        pop     ax                      ;
        ret                             ;

;----------------------------------------------------------------
; Routine GetLock
; purp: Get a read/write lock bit for a unit
; in  : al = unit number
; out : carry = read/write lock bit for that unit
; uses: flags

GetLock:push    ax                      ; save input
        push    cx                      ;

        ; test high/low word
        cmp     al,16                   ;
        jnc     getuplk                 ; upper byte lock bit

        ; lock bits 0 .. 15 in lower word
        mov     cl,al                   ; make cl = bit # in ax
        mov     ax,word ptr LockBits    ; get lockbits word
        jmp     getlkbit                ;

        ; lock bits 16 .. 31 in upper word
getuplk:sub     al,16                   ; make cl = bit # in ax
        mov     cl,al                   ;
        mov     ax,word ptr LockBits+2  ; get lockbits word
;       jmp     getlkbit                ;

        ; get the correct bit to the carry
getlkbit:
        shr     ax,cl                   ; get the bit
        and     ax,1                    ; mask all other bits
        add     ax,0FFFFH               ; put in the carry

        pop     cx                      ;
        pop     ax                      ;
        ret                             ;

;End------------------------------------------------------------

;Defs-----------------------------------------------------------
;
; CF interface defines
;
; All defines of parameters that are not bytes etc..
; These take no memory, thay are only symbolic names for things I
; intend to use later. Mainly the defines etc. of the CF
; register parameters.
;
; The I/O ports of the CF.
;
IoBase          equ     0F000H  ; I/O base address for the interface
;
; CF register adresses. The CF 'bus' looks like these I/O ports
; to an outside observer.
;
CfDevCtl       equ     0EH+IoBase      ; device control register
CFCmd          equ     07H+IoBase      ; addres for command
CFSts          equ     07H+IoBase      ; addres status register
CFLBA3         equ     06H+IoBase      ; addres LBA3
CFLBA2         equ     05H+IoBase      ; addres LBA2
CFLBA1         equ     04H+IoBase      ; addres LBA1
CFLBA0         equ     03H+IoBase      ; addres LBA0
CFnum          equ     02H+IoBase      ; addres number of sectors
CFErr          equ     01H+IoBase      ; addres Error register
CFFeat         equ     01H+IoBase      ; features register
CFData         equ     00H+IoBase      ; addres for data bus
;
; The LBA3 register also contains the masks for LBA modus, and
; the master/slabe bit. I fixed this at master device, I have no
; real expectations of anyone using this thing with *two* CF's
; connected to one interface.
;
CFLBA3AM       equ     00001111B       ; LBA3 AND and OR masks
CFLBA3OM       equ     11100000B       ;
;
; The Reset/IRQ register has two interesting bits. Later I'll
; have to implement a soft-reset of the disk in case of errors.
; At this moment I do not use these register bits.
;
IDESRes         equ     00000100B       ; Soft Reset bit
IDENIRQ         equ     00000010B       ; 0 = IRQ active
;
; My usage of the soft-reset bits, I keep the interrupts off
;
CFSetSRes       equ     00000110B       ; set soft-reset
CFClrSRes       equ     00000010B       ; clear soft-reset
;
; The bits from CF Status register
;
StsBsy          equ     10000000B       ; Busy flag
StsRdy          equ     01000000B       ; Ready flag
StsWft          equ     00100000B       ; Write error
StsSKC          equ     00010000B       ; Seek complete
StsDRQ          equ     00001000B       ; Data Request
StsCorr         equ     00000100B       ; ECC executed
StsIdx          equ     00000010B       ; Index found
StsErr          equ     00000001B       ; Error flag
;
; Command opcodes I use
;
CmdRead         equ     020H    ; Read a sector
CmdWrite        equ     030H    ; Write a sector
CmdIdent        equ     0ECH    ; Identify disk
CmdFeat         equ     0EFH    ; Set features
FeatPower       equ     09AH    ; Power modus feature
;
;End-------------------------------------------------------------------

;Static---------------------------------------------------------
;
; The static storage of the IDE interface driver.
;
LBB     dd      0       ; LBA base (Partition start)
LBA     dd      0       ; Lineair Block Address (LBA)
SCnt    dw      0       ; sector count.
idnoname    db      0   ; no name specified
;
; the parameters of the disk (as read from the disk using the
; ident command)
;
; flag, used to remember if a name was specified of the disk.
;
SPD     dd      0       ; Sectors Per Disk
;
;End------------------------------------------------------------

;Routines-------------------------------------------------------
;
; The CF I/O routines.
;
;---------------------------------------------------------------
; Routine DoCmd         CFok
; purp: issues command opcode to the IDE device
; in  : nothing
; out : nothing
; uses: al

DoCmd:
        mov     dx,CFCmd       ; set address
        
;---------------------------------------------------------------
; Routine WriteByte     CfOk
; purp: Write one byte to the IDE bus
; in  : byte in AL
; out : nothing
; uses: nothing

WriteByte:
        out     dx,al           ;
        ret                     ;

;---------------------------------------------------------------
; Routine ReadSts       CFOk
; purp: get status if the IDE device
; in  : nothing
; out : status byte in al
; uses: nothing

ReadSts:
        mov     ax,CfSts        ; set address

;---------------------------------------------------------------
; Routine ReadReg       CFOk
; purp: get byte from the IDE device
; in  : address in ax
; out : data byte in al
; uses: nothing

ReadReg:mov     dx,ax           ;

;---------------------------------------------------------------
; Routine ReadByte      CfOk
; purp: Read one byte from the IDE bus
; in  : dx = I/O address
; out : byte in al
; uses: ax
; nb  : assumes address and bus have been set

ReadByte:
        in      al,dx           ;
        ret                     ;

;---------------------------------------------------------------
; Routine WaitNBSY      CFOk
; purp: wait till the drive indicates non-busy status
; in  : nothing
; out : errorc and carry set if timed out
; uses: nothing

WaitNBSY:

        push    ax              ; save registers
        push    cx              ;
        push    dx              ;
        mov     cx,0            ; timeout

        ; test if device ready
wtrdy1: call    ReadSts         ; get status byte
        and     al,StsBsy       ; get status bits
        jz      to_FinWait      ;
        loop    wtrdy1          ; wait for it

        jmp     WaitTimOut      ;

to_FinWait:jmp  FinWait         ;

;---------------------------------------------------------------
; Routine WaitDRDY      CFOk
; purp: wait till the drive indicates Data ready status
; in  : nothing
; out : errorc and carry set if timed out
; uses: nothing

WaitDRDY:

        push    ax              ; save registers
        push    cx              ;
        push    dx              ;
        mov     cx,0            ; timeout

        ; test if device ready
wtdrdy1:call    ReadSts         ; get status byte
        and     al,StsRdy       ; get status bits
        jnz     FinWait         ;
        loop    wtdrdy1         ; wait for it

        jmp     WaitTimOut      ;

;---------------------------------------------------------------
; Routine WaitDrq
; purp: wait till the drive indicates data request status
; in  : nothing
; out : errorc and carry set if timed out
; uses: nothing

WaitDrq:

        push    ax              ; save registers
        push    cx              ;
        push    dx              ;
        mov     cx,0            ; timeout

        ; test if device ready for data
wtdrq1: call    ReadSts         ; get status byte
        test    al,StsDrq       ; get status bits
        jnz     FinWait         ;
        loop    wtdrq1          ;

        jmp     WaitTimOut      ;

;---------------------------------------------------------------
; Fragment FinWait
; purp: finishes up things when waiting for something is done
; in  : ax still on stack
; out : nothing
; uses: nothing
; NB  : this code fragment is a result of code compression
;       it is used by the routines WaitDRQ, WaitNBSY and
;       WaitDRDY

FinWait:pop     dx              ;
        pop     cx              ; get cx back from the stack
        pop     ax              ; get ax back from stack
        clc                     ; no errors
        ret                     ; routine done

;---------------------------------------------------------------
; Fragment WaitTimOut
; purp: signals timeout errors
; in  : cx,ax still on stack
; out : nothing
; uses: nothing
; NB  : this code fragment is a result of code compression
;       it is used by the routines WaitDRQ, WaitNBSY and
;       WaitDRDY.
; NB1 : A soft-reset added to make the disk non-lockup if some
;       timeout happens
;

WaitTimOut:

        mov     errorc,ENRDY    ; set error code

        ; get registers from stack
        pop     dx              ;
        pop     cx              ;
        pop     ax              ; get ax back from stack
        stc                     ; signal error
        ret                     ; routine done

;---------------------------------------------------------------
; Routine SoftReset CFok
; purp: issue a soft-reset to the disk
; in  : nothing
; out : nothing
; uses: nothing

SoftReset:
        push    ax              ; save registers
        push    dx              ;
        push    cx              ;

        ; Issue a soft-reset to the disk, perhaps it can recover
        ; from errors this way...
        mov     dx,CFDevCtl     ; set the address
        mov     al,CFSetSRes    ; set soft-reset
        out     dx,al           ; set the reset pulse

        mov     cx,100H         ;
wtout1: loop    wtout1          ;

        mov     ax,CFClrSRes    ; remove soft-reset
        out     dx,al           ;

        mov     cx,100H         ;
wtout2: loop    wtout2          ;

        ; wait till busy bit cleared
        mov     cx,0080H        ; wait long tim
wtout3: call    ReadSts         ; get status byte
        test    al,StsBsy       ; get status bits
        jz      wtout4          ;
        loop    wtout3          ; wait for it
wtout4:

        ; restore registers
        pop     cx              ;
        pop     dx              ;
        pop     ax              ;

        ret                     ;

;---------------------------------------------------------------
; Routine SetLBA
; purp: sets the LBA for the next transfer
; in  : LBA holds sector number
; out : registers of the IDE controller loaded
; uses: CCyl, CHead and CSec set
; NB  : I select ONE sector to be processed

SetLBA:

        ; save register ax
        push    ax                      ;
        push    dx                      ;

        ; prepare for CF registers

        ; use real LBA access
        mov     ax,word ptr LBA         ;

        mov     dx,CFLBA0               ;
        out     dx,al                   ;

        mov     al,ah                   ;
        mov     dx,CFLBA1               ;
        out     dx,al                   ;

        mov     ax,word ptr LBA+2       ;

        mov     dx,CFLBA2               ;
        out     dx,al                   ;

        mov     al,ah                   ;
        and     al,CFLBA3AM             ; mask for LBA, device 0
        or      al,CFLBA3OM             ;
        mov     dx,CFLBA3               ;
        out     dx,al                   ;

        mov     al,1                    ; always 1 sector
        mov     dx,CFNum                ;
        out     dx,al                   ;

        pop     dx                      ; get dx back
        pop     ax                      ; get ax back

        ret                             ; done

;---------------------------------------------------------------
; Fragments errors1     CFok
; purp: provide jump part to the ReadSecErr etc..
;       They are out of reach here
; in  : nothing
; out : nothing
; uses: nothing

ReadSecErr1:                    ;
        jmp     ReadSecErr      ;
ReadSecStsErr1:                 ;
        jmp     ReadSecStsErr   ;
ReadSecDrqErr1:                 ;
        jmp     ReadSecDrqErr   ;

;---------------------------------------------------------------
; Routine RdSkip        CFok
; purp: skip data reads from the disk
; in  : cx = number of words to skip
; out : nothing
; uses: ax

RdSkip: call    ReadWord        ; read a word
        loop    RdSkip          ; ignore it
        ret                     ;

;---------------------------------------------------------------
; Routine ReadWord      Cfok
; purp: reads data from the disk
; in  : all setup has to be done already
; out : ax = word read
; uses: dx

ReadWord:

        mov     dx,CFData       ; get data word
        in      al,dx           ;
        mov     ah,al           ;
        in      al,dx           ;
        xchg    al,ah           ;

        ret                     ;

;---------------------------------------------------------------
; Routine ReadSec       Cfok
; purp: reads one sector from the disk
; in  : LBA   = block address to write
;       es:di = address to store the data
; out : data read into the destination address, carry set if error
; uses: nothing

        ; wait for bsy bit cleared
ReadSec:
        call    WaitNBSY        ; wait till disk is ready
        jc      ReadSecErr      ;
        call    WaitDRDY        ;
        jc      ReadSecErr      ;

        ; load parameters
        call    SetLBA          ; set address
        mov     al,CmdRead      ; give read command
        call    DoCmd           ;

        ; wait for busy gone
        call    WaitNBSY        ;
        jc      ReadSecErr      ;

        ; test on errors
        call    ReadSts         ; get status byte
        test    al,StsErr       ; check error bit
        jnz     ReadSecStsErr   ;

        ; test if ready for data transport
        call    WaitDRQ         ;
        jc      ReadSecDrqErr   ;

        ; transport the data block to destination
        call    ReadBlock       ; get the data block

        ; final status check
        call    ReadSts         ; get status byte
        test    al,StsErr       ; check error bit
        jnz     ReadSecStsErr   ;

        ; return with no errors
        clc                     ; no errors
        ret                     ;

        ; DRQ error, no DRQ after command passed
ReadSecDrqErr:                          ;
        mov     errorc,EREAD            ; read error

        ; errorc is already set elsehere
ReadSecErr:
        call    SoftReset               ; soft-reset to the disk
        stc                             ;
        ret                             ;

ReadSecStsErr:
        mov     errorc,EREAD            ; preset to read error
;       jmp     RWStsErr                ; see below...

;---------------------------------------------------------------
; Fragment RWStsErr     Cfok
; purp: attempts to get a proper error code after read/write
;       sector failed.
; in  : byte errorc preset to an error code
; out : sometimes a better error code in errorc
; uses: nothing
; NB  : this code is used by readsec and writesec

        ; status gives error code
        ; see if I can make out any error remark
RWStsErr:
        mov     ax,CFErr                ; get error register
        call    ReadReg                 ;

        ; see if I can make something of this
        test    al,00010001B            ; sector not found
        jz      stserr1                 ; addres mark not found
        mov     errorc,ESECNF           ; gives sector not found
        jmp     StsErrDone              ;

stserr1:

        test    al,01000000B            ; uncorrectable
        jz      stserr2                 ; data error
        mov     errorc,ECRC             ; gives CRC error
        jmp     StsErrDone              ;

stserr2:

        test    al,00000010B            ; track 0 not found
        jz      stserr3                 ;
        mov     errorc,ESEEK            ; gives seek error
        jmp     StsErrDone              ;

stserr3:

        test    al,00000100B            ; aborted
        jz      stserr4                 ;
        mov     errorc,EGENERAL         ; gives general failure
        jmp     StsErrDone              ;

stserr4:

        ; not anything I know how to handle, use default and exit

StsErrDone:

        call    SoftReset       ; soft-reset the disk
        stc                     ; set error flag
        ret                     ; exit

;---------------------------------------------------------------
; Routine ReadBlock     CFok
; purp: read one block from the IDE device
; in  : es:di -> request packet
; out : es:di 256 words further on
; uses: nothing

ReadBlock:
        ; save registers
        push    ax              ;
        push    cx              ;
        push    dx              ;

        ; setup for data read
        mov     dx,CFData       ; set address

        ; init loop counter
        mov     cx,512          ; 512 bytesreads
        cld                     ; clear direction flag (UP)

        ; setup source pointer
        les     di,[DataPointer]        ;

rdblk1: mov     dx,CFData       ; use word read
        in      al,dx           ;
        stosb                   ; put in memory

        ; do for all words of a block
        loop    rdblk1          ;

        ; make pointers ok again
        mov     word ptr DataPointer,di     ; store new
        mov     word ptr DataPointer+2,es   ; pointer

        les     di,[DosPointer] ; get old es:di back

        ; done, get things back
        pop     dx              ;
        pop     cx              ;
        pop     ax              ;
        ret                     ;

;---------------------------------------------------------------
; Routine WriteSec      Cfok
; purp: Write one sector to the disk
; in  : LBA   = sector to write to
;       es:di -> request packet
; out : data written to the disk, carry set if error happened
; uses: nothing

WriteSec:
        call    WaitNBSY        ; wait till bus is ready
        jc      WriteSecErr     ;
        call    WaitDRDY        ; wait till disk is ready
        jc      writeSecErr     ;

        call    SetLBA          ; set address
        mov     al,CmdWrite     ; give write command
        call    DoCmd           ;

        ; wait till command ready
        call    WaitNBSY        ;
        jc      WriteSecErr     ;

        ; test on errors
        call    ReadSts         ; get status byte
        test    al,StsErr       ; check error bit
        jnz     WriteSecStsErr  ; error -> report

        ; wait for DRQ
        call    WaitDRQ         ;
        jc      WriteSecDrqErr  ;

        ; transport the data block to destination
wrsec1: call    WriteBlock      ; write data to disk

        ; test final status
        call    ReadSts         ;
        test    al,StsErr       ; check error bit
        jnz     WriteSecStsErr  ; error -> report

        ; return indicating no errors
        clc                     ;
        ret                     ; done

        ; No Drq after command given
WriteSecDrqErr:                         ;
        mov     errorc,EWRITE   ; write fault

        ; errorc is set elsewhere
WriteSecErr:                    ;
        call    SoftReset       ; give disk a soft reset
        stc                     ; error indicator on
        ret                     ;

        ; error status detected
WriteSecStsErr:                         ; write fault too
        mov     errorc,EWRITE           ; default = write fault
        jmp     RWStsErr                ; perhaps there is a
                                        ; better one

;---------------------------------------------------------------
; Routine WriteBlock    Cfok
; purp: write one block of data to the IDE device
; in  : es:di -> data address (512 bytes)
; out : nothing
; uses: nothing

WriteBlock:

        ; save registers
        push    ax              ; accu
        push    cx              ; 256 words counter
        push    dx              ; I/O port address
        push    ds              ; save ds:si
        push    si              ; on stack this time..

        ; setup for data write
        mov     dx,CFData       ; set address

        ; here I now do the es:di -> ds:si transit
        lds     si,[DataPointer]        ;

        ; init loop counter
        mov     cx,512          ; 512 byteswrite
        cld                     ; direction = up

wrblk1: lodsb                   ; get data w
        out     dx,al           ;

        ; do that for all 512 bytes
        loop    wrblk1          ;

        ; put new data destination pointer back
        mov     word ptr cs:DataPointer,si      ;
        mov     word ptr cs:DataPointer+2,ds    ;

        ; done, get things back
        pop     si              ; ds:si from stack
        pop     ds              ;
        pop     dx              ; registers too
        pop     cx              ;
        pop     ax              ;
        ret                     ;

;End------------------------------------------------------------

;End------------------------------------------------------------

; **********************************************
; * End of the resident part of the driver     *
; * all code PAST this label will not be there *
; * after the Init routine has been called.    *
; **********************************************

EndResident:

;---------------------------------------------------------------
; Routine ParsePars
; purp  : reads and interpretes the driver's command line
; in    : request block data = command line
; out   : driver's parameters set, carry set if something wrong
;         carry clear if parsing went ok.
; uses  : nothing

ParsePars:

        ; I'll use ds:si to scan the command line. that means
        ; that ds no longer points to the local code segment
        push    ds                      ; save ds for later

        ; make ds:si -> command line
        mov     si,word ptr es:[di+RqInitCmdLine]  ; offset command line
        mov     ax,word ptr es:[di+RqInitCmdLine+2]; segment command line
        mov     ds,ax                   ;

        ; I'll use string operations -> direction = up.
        cld                             ;

Parameter:

        ; scan the command-line
        lodsb                   ; get a command-line byte
        cmp     al,01FH         ; test for end-of-line
        ja      ParamH          ; look for parameter head
        jmp     ParamEnd        ; <CR> ends all parameter scans

        ; see if a parameter starts here
ParamH: cmp     al,'/'          ; parameters start with '/'
        jz      Param           ;
        cmp     al,'-'          ; or with '-' (UNIX-style)
        jz      Param           ;
        jmp     Parameter       ;

        ; parse parameters
Param:  lodsb                   ; get next byte
        and     al,0DFh         ; convert to UPPER case

        ; Parameter 'D': use debug modus
        cmp     al,'D'          ; test debug modus
        jne     nota_D          ;

        inc     cs:[debug]      ;
        jmp     Parameter       ;

nota_D:

        ; Parameter 'N': no file-system check
        cmp     al,'N'          ; file-system check skip
        jne     nota_N          ;

        inc     cs:[NoFsChk]    ;
        jmp     Parameter       ;

nota_N:

        ; none of the parameters I know, give parameter
        ; error

ParamError:

        pop     ds                      ; get ds back

        mov     dx,Offset CmdErrorStr   ; Command-line error
        mov     ah,9                    ;
        int     21h                     ;

        stc                             ; carry set
        ret                             ;

        ; parameter parsing worked out ok.

ParamEnd:

        pop     ds                      ; restore data segment
        clc                             ;

        ret

;---------------------------------------------------------------
; Routine Number
; purp: reads a number from a <CR> terminated string
; in  : ds:si -> string
; out : ax = number's value, carry clear when number read or
;       ax = 0000, carry set when error happened.
;       both cases: si -> last byte that was not understood or
;                   not part of the number.
; uses: si

Number: push    cx              ; save cx, is result register
        push    dx              ; save dx, needed for mul
        mov     cx,0            ;

        ; start the scan
numb0:  lodsb                   ; get the first byte
        cmp     al,13           ; see if line ends
        ja      numb1           ;

        ; end of line, si -> cr
numerr: dec     si              ; si -> the cr
        mov     ax,0            ; result = 0000
        pop     dx              ;
        pop     cx              ; cx from stack
        stc                     ; error signal set
        ret                     ; done

        ; next test
numb1:  cmp     al,' '          ; separation is ' ' or '/'
        jz      numb0           ; many ' '
        cmp     al,'/'          ;
        jz      numb3           ; or a single '/'

        ; now the real work begins.
        sub     al,'0'          ;
        jc      numerr          ; below '0' is no good
        cmp     al,10           ;
        jnc     numerr          ; above '9' is no good too

        ; found a digit in al
numb2:  push    ax              ; I want: mul cx,10
        mov     ax,cx           ; have to move about somewhat
        mov     cx,10           ;
        mul     cx              ;
        mov     cx,ax           ;
        pop     ax              ;

        add     cl,al           ; put result in place
        adc     ch,0            ;

numb3:  lodsb                   ; get next byte
        sub     al,'0'          ; test it
        jc      numfin          ;
        cmp     al,10           ;
        jnc     numfin          ;

        jmp     numb2           ; process it

        ; finished reading the number
numfin: mov     ax,cx           ; get result
        dec     si              ; si -> non-number
        pop     dx              ;
        pop     cx              ;
        clc                     ;
        ret                     ;

;---------------------------------------------------------------
; Routine CheckIO
; purp: checks if the actual computer is a PortFolio, and if the
;       drive/interface is ok.
; in  : nothing
; out : # of drives set to 0 if anything goes wrong.
; uses: a lot, registers; 82C55 etc....
;---------------------------------------------------------------

CheckIO:

        ;--------------------------------------------------------
        ; test if this is a PoFo (in fact: if it's NOT a normal PC)
        ;--------------------------------------------------------
        in      al,61h                  ; read keyboard port of PC
        cmp     al,61h                  ; gives 61H in a PoFo
        jz      isa_PF                  ;

        ; prepare for error message
        mov     dx,Offset NoPoFoStr     ;
        jmp     CheckFailedStr          ;

        ; PoFo detected, continue installation
isa_PF:

        ;--------------------------------------------------------
        ; init CF module
        ;--------------------------------------------------------
        call    InitCf                  ; init bus and reset drive
        jnc     ICFDone                 ; if no device found

        ; InitCf returned an error
        jmp     CheckFailedStr          ;

ICFDone:

        ;--------------------------------------------------------
        ; Get the module's size
        ;--------------------------------------------------------

        ; do the Ident part
        call    WaitNBSY        ; wait till disk ready
        jc      IdentFailed     ; may time out...
        call    WaitDRDY        ;
        jc      IdentFailed     ;

        ; give command
        mov     al,CmdIdent     ; send Ident command
        call    DoCmd           ;

        ; wait for busy gone
        call    WaitNBSY        ;
        jc      IdentFailed     ; timeout?
        call    WaitDRDY        ; and data-ready there
        jc      IdentFailed     ; timeout?

        ; test on errors reported by the drive itself
        call    ReadSts         ; get status byte
        test    al,StsErr       ; check error bit
        jnz     IdentFailed     ; no error -> cont

        ; device ready for data transfer
        call    WaitDRQ         ; drive must make DRQ active
        jc      Identfailed     ; or else I do not know what...

        ; setup for data read
        mov     dx,CFData       ; dx -> data port
        jmp     idcont          ;

        ; I need a jmp to get to Check_Failed
IdentFailed:                            ; otherwise: jmp out of
        mov     dx,offset NoIdentStr    ;
        jmp     CheckFailedStr          ; range

        ; I use cx as word counter
idcont: mov     cx,0            ; no words read yet

        ; read a word from the disk
idloop: call    ReadWord        ;


        ; test if config byte is there
        cmp     cx,0            ;
        jnz     idloop1         ;

        cmp     ax,0848AH       ;
        jnz     IdentFailed     ;

idloop1:
        ; test if # sectors per disk MSW
        cmp     cx,7            ;
        jnz     idloop2         ;

        mov     word ptr SPD+2,ax       ;

idloop2:cmp     cx,8            ;
        jnz     idloop3         ;

        mov     word ptr SPD,ax         ;

idloop3:

        ; just before the name of the device
        cmp     cx,26           ; just before the name
        jnz     idloop4         ;

        mov     dx,offset IdeNameStr
        mov     ah,9            ;
        int     21h             ;
        jmp     idloope         ;

idloop4:cmp     cx,27           ; word # 27 = first word of
        jnz     idloop5         ;             device name

        ; test if name there
        cmp     ax,0            ;
        jnz     idloop5         ; name is there
        mov     idnoname,1      ;
        mov     dx,offset IdeNoName     ;
        mov     ah,9                    ;
        int     21h                     ;

        ; test if in the name
idloop5:cmp     cx,27           ; words # 27 .. 46 = disk type
        jl      idloope         ; ASCII string
        cmp     cx,40           ; I stop at word # 40
        jg      idloope         ;

        ; test if name there
        cmp     idnoname,1      ;
        jz      idloope         ;

        ; inside the name
        xchg    ah,al           ; print the bytes
        call    DisChar         ; high-low, in that order
        xchg    ah,al           ;
        call    DisChar         ;

        ; end of checks
idloope:nop                     ;

        ; do for all words of a block
        inc     cx              ;
        cmp     cx,256          ;
        jnz     idloop          ;

        ; end of disk ident result handling
        call    NewLine         ; for disk type string output

        ;--------------------------------------------------------
        ; compute the number of units that this disk can contain
        ;--------------------------------------------------------
        mov     Units,0                 ; start with 0 units
        mov     ax,word ptr SPD         ;
        mov     dx,word ptr SPD+2       ;

        ; compute the number of full partitions that will fit
        ; nb: dx:ax = SPD

        ; get the number of full
        mov     cx,dsize                ;
        div     cx                      ;
        mov     Units,al                ;

        ; see how many you got
        cmp     ax,23                   ; less than 23 gives some
        jc      unit1                   ; work
        jz      NoTail                  ; 23, means a full house

        ; more than a full disk, set to 23 units and exit
        mov     Units,23                ;
        jmp     NoTail                  ;

        ;--------------------------------------------------------
        ; a number of tail blocks is there, see what to do
        ; with them. This is a kind of up-and-down calculation.
        ; First I compute the best cluster size, then the number
        ; of FAT sectors, then I re-generate the disk size I
        ; would get from that. I fill in the parameters as I go
        ; along.
        ;--------------------------------------------------------
unit1:  mov     ax,dx                   ; get remainder

        ; see if it any use to bother about this tail
        cmp     ax,(minsize+admsize)    ; if too small
        jc      notail                  ; I skip it

        ; prepare tail's parameters. This is some kind of
        ; 'best-fit' of a file-system into the remaining
        ; file system parameters.

        ; determine the best sectors/cluster size
        push    ax                      ; save ax
        mov     cx,4096                 ; max no of data clusters
        mov     dx,0                    ;
        div     cx                      ;
        cmp     dx,0                    ; if it fits...
        jz      unit2                   ; no rounding, else
        inc     al                      ; round up
unit2:  mov     ClustL,al               ;
        pop     ax                      ; get ax back

        ; got the cluster size, put parameters in place
fitone: mov     cl,ClustL               ; see how many clusters
        cmp     cl,0                    ; I NEVER div/0
        jz      NoTail                  ;

        mov     DSizeL,ax               ; put that in place too
        mov     ax,12                   ;
        mov     FatSizeL,ax             ; ax = # of sectors FAT

        ; fitted the tail into the FAT12 file-system one more
        ; unit done
        inc     Units                   ;
        jmp     UntDone                 ;

        ; no tail partition, last partition has size of
        ;  a normal partition
NoTail:
        mov     word ptr DSizeL,dsize           ;
        mov     byte ptr ClustL,cluster         ;
        mov     word ptr FatSizeL,fatsize       ;

UntDone:

        ;--------------------------------------------------------
        ; test the file system info of disk
        ;--------------------------------------------------------
        cmp     NoFsChk,0               ; test if fs check is
        jnz     SkipFsCheck             ; bypassed

        ; execute fs check
        mov     CUnit,0                 ; for all units

        ; test file-system there
testfslp:

        mov     al,CUnit                ; setup unit number
        call    TestFs                  ; test fs
        jnc     tstfsok                 ; test for errors

        ; error processing
        mov     dx,offset FsErrStr      ; display error message
        mov     ah,9                    ;
        int     21h                     ;
        mov     al,CUnit                ;
        add     al,'d'                  ; with drive letter
        call    DisChar                 ;
        mov     al,':'                  ; and ':'
        call    DisChar                 ;
        call    NewLine                 ;

        mov     al,CUnit                ; lock this unit
        call    SetLock                 ;

tstfsok:inc     CUnit                   ; next fs
        mov     al,CUnit                ;
        cmp     al,Units                ;
        jnz     testfslp                ;

        mov     dx,offset FsDnStr       ;
        jmp     FsChkDone               ;

SkipFsCheck:
        mov     dx,offset NoFsStr       ;

FsChkDone:
        ; in debug modus: bark
        cmp     debug,0         ;
        jz      FsNoBark        ;

        ; give message
        mov     ah,9            ;
        int     21h             ;

FsNoBark:

        ;--------------------------------------------------------
        ; all checks passed
        ;--------------------------------------------------------
Check_ok:

        ; tell the user what drives to expect
        mov     dx,offset DrvStr; d: is always there
        mov     ah,9            ;
        int     21h             ;

        ; test if E: etc there
        cmp     Units,1         ;
        jz      InitDrvDone     ;
        mov     dx,offset DrvStr1       ;
        mov     ah,9            ;
        int     21h             ;

        mov     dl,Units        ;
        add     dl,'d'-1        ;
        mov     ah,2            ;
        int     21h             ;

        mov     dx,offset DrvStr2       ;
        mov     ah,9            ;
        int     21h             ;

InitDrvDone:

        mov     dx,offset NewLn ;
        mov     ah,9            ;
        int     21h             ;

        clc                             ;
        ret                             ;

        ;--------------------------------------------------------
        ; process errors
        ; dx already points to an error message
        ;--------------------------------------------------------

CheckFailedStr:                         ;

        mov     units,1                 ; set number of units to
                                        ; 1 to prevent hangs
        mov     ah,9                    ; issue error message
        int     21h                     ;

        ;--------------------------------------------------------
        ; process errors
        ; Error message already given
        ;--------------------------------------------------------


CheckFailedNoStr:

        stc                             ;
        ret                             ;

;---------------------------------------------------------------
; Routine InitCf
; purp: Init the CF module
; in  : nothing
; out : carry set if error happened
; uses: nothing

        ; set I/O port modus to activate the control signals
InitCf:

        mov     dx,CFLBA0       ; test if CF is there
        mov     al,055H         ; check on CF LBA0 register
        out     dx,al           ;
        nop                     ;
        nop                     ;
        in      al,dx           ;
        cmp     al,055H         ;
        jnz     NoCf            ;

        mov     al,0AAH         ;
        out     dx,al           ;
        nop                     ;
        nop                     ;
        in      al,dx           ;
        cmp     al,0AAH         ;
        jnz     NoCf            ;

        ; set the CF to low power modus
        call    WaitNBSY        ;
        mov     dx,CFNum        ;
        mov     al,1            ;
        mov     dx,CfFeat       ;
        mov     al,FeatPower    ;
        out     dx,al           ;
        mov     dx,CfCmd        ;
        mov     al,CmdFeat      ;
        out     dx,al           ;
        jmp     InitCfOk        ;

        ; CF not there
NoCf:   mov     dx,offset NoCfFound ; these two bits are defined '1'
        jmp     InitCFErr           ; and reserved at that value. If they
                                    ; are not there, the device is not there

        ; ok return of InitCf, carry is clear here
InitCFOk:
        mov     errorc,0        ;
        clc                     ;
        ret                     ;

        ; error processing of the InitCf routine
InitCFErr:
        mov     errorc,0        ; INIT does NOT accept errors
        stc                     ; carry set
        ret                     ; and return

;---------------------------------------------------------------
; Routine TestFs
; purp: check the BPB from the disk
; in  : al = unit no to test the file-system for
; out : carry clear if file-system there, else carry set
; uses: registers

TestFs:

        ; prepare buffers
        call    FixBPB                  ; make reference BPB for unit
        call    SetLBB                  ; make LBB for unit

        ; prepare parameters for a modified read sector
        mov     ax,word ptr LBB         ; read sector 0 of current
        mov     word ptr LBA,ax         ; unit
        mov     ax,word ptr LBB+2       ;
        mov     word ptr LBA+2,ax       ;

        ; wait for bsy bit cleared
        call    WaitNBSY        ;
        jc      TstFs5          ;
        call    WaitDRDY        ;
        jc      TstFs5          ;

        ; load parameters
        ; set this LBA in the disk
        call    SetLBA          ; set address

        ; issue read command
        mov     al,CmdRead      ;
        call    DoCmd           ;

        ; wait for busy gone
        call    WaitNBSY        ;
        jc      TstFs5          ;

        ; test on errors
        call    ReadSts         ; get status byte
        test    al,StsErr       ; check error bit
        jnz     TstFs5          ;

        ; wait till drive is ready for data
        call    WaitDRQ         ; wait for drq
        jc      TstFs5          ;

        ; get the bootblock into the test buffer
        push    ax              ;
        push    cx              ;
        push    dx              ;
        push    di              ;
        push    es              ;

        ; setup for data read
        mov     dx,CFData       ;

        ; get BPB data words
        mov     cx,20           ; read 20 words (40 bytes)

        cld                     ; read into buffer
        mov     ax,cs           ; testblock
        mov     es,ax           ;
        mov     di,offset TestBlock

tstfs1: call    ReadWord        ;
        stosw                   ;
        loop    TstFs1          ;

        ; skip the rest
        mov     cx,246          ; skip the tail of the block
        call    RdSkip          ;

        ; see if the Fs from the disk matches the one in BootBlock
        ; issue warning if not so
        ; I'll test all the disk-config parameters
        mov     si,offset BootBlock     ; BPB as reference
        mov     di,offset Testblock     ; TST to be tested
        mov     cx,bootsize             ; # of bytes

        ; go test if the same
tstfs2: mov     al,[si]                 ;
        cmp     al,[di]                 ;
        jnz     tstfs3                  ;
        inc     si                      ;
        inc     di                      ;
        loop    tstfs2                  ;

        ; all testable parameters are ok, hope for the best
        ; about the rest of the parameters
        clc                     ; error flag off
        jmp     tstfs4          ;

        ; registers back, error exit
tstfs3: stc                     ; error flag on
tstfs4: pop     es              ;
        pop     di              ;
        pop     dx              ;
        pop     cx              ;
        pop     ax              ;
        ret                     ;

tstfs5: stc                     ;
        ret                     ;

;
;---------------------------------------------------------------
; I use a seccond buffer to load the header of the file-system
; when testing it. This costs only a few words of memory space,
; saves a load of counting however. The thing is a copy
; of the real bootblock.
;---------------------------------------------------------------
;
TestBlock:
        jmp     short notest    ; short jmp ends up here
        db      0               ; filler byte
TstNam  db      "        "      ; volume name (8 chars)
TST     dw      0          ;+00 ; sector size     (fixed: 0200H)
TSTclus db      0          ;+02 ; cluster size    (fixed: 16)
TSTResv dw      0          ;+03 ; reserved sectors (mostly 1)
TSTfatn db      0          ;+05 ; no of FAT's     (fixed: 1)
TSTroot dw      0          ;+06 ; root dir entries
TSTdsiz dw      0          ;+08 ; total number of sectors
TSTdtyp db      0          ;+0A ; media code      (fixed: F8H = hard-disk)
TSTfats dw      0          ;+0B ; sectors per FAT (fixed: 12)
        dw      0          ;+0D ; sectors per track (nonsense
        dw      0          ;+0F ; number of heads     really)
        dd      0          ;+11 ; hidden sectors
        db      0,0,0,0,0       ; reserved
        db      0,0,0,0,0,0     ; reserved too
;
notest:
;
; For comparison, I need to know the size of the testblock.
; crazy MASM: gives phase-errors on this... use rootsize
; instead. Gives the same number, but no phase errors....
;
testsize        equ (offset notest - offset testblock)
;
;---------------------------------------------------------------------
; Routine NewLine
; purp: make the console goto a new line
; in  : nothing
; out : nothing
; uses: nothing

Newline:push    ax              ; save registers
        push    dx              ;
        push    ds              ;

        mov     ax,cs           ; ds <- cs
        mov     ds,ax           ;

        mov     ah,9            ; print predefined string
        mov     dx,Offset NewLn ;
        int     21h             ;

        pop     ds              ; get registers back
        pop     dx              ;
        pop     ax              ;

        ret                     ; done

;--------------------------------------
; Routine DisHexByte and DisHexWord
; purp  : Writes AL or AX as hex ascii
; in    : byte in AL (DisHexByte)
;         word in AX (DisHexWord)
; out   : nothing
; uses  : nothing

DisHexWord:

        ; save registers
        push    ax              ;

        ; higher byte
        mov     al,ah           ;
        call    DisHexByte      ;

        ; lower byte
        pop     ax              ;
        call    DisHexByte      ;

        ; done
        ret                     ;

DisHexByte:

        ; save registers
        pushf                   ;
        push    ax              ;

        ; higher nibble
        shr     al,1            ;
        shr     al,1            ;
        shr     al,1            ;
        shr     al,1            ;
        call    ToHex           ;
        call    DisChar         ;

        ; lower nibble
        pop     ax              ;
        push    ax              ;
        and     al,0FH          ;
        call    ToHex           ;
        call    DisChar         ;

        ; restore registers
        pop     ax              ;
        popf                    ;

        ; done
        ret                     ;

; --------------------------------------
; Routine ToHex
; purp  : converts nibble to hex ASCII character
; in    : nibble in al
; out   : ASCII char in al
; uses  : nothing

ToHex:  pushf                   ;

        and     al,0FH          ; make sure of 0 .. F

        cmp     al,0AH          ;
        jge     tohex1          ;

        ; 0 .. 9
        add     al,'0'          ; 0 .. 9 convert
        jmp     tohexe          ;

        ; A .. F
tohex1: add     al,('A' - 0AH)  ;

tohexe: popf                    ;

        ret                     ;

; --------------------------------------
; Routine DisChar
; purp  : Writes AL to display
; in    : character in al
; out   : nothing
; uses  : nothing

DisChar:pushf                   ;
        push    ax              ;
        push    dx              ;

        mov     dl,al           ;
        mov     ah,02H          ;
        int     21H             ;

disnsio:pop     dx              ;
        pop     ax              ;
        popf                    ;

        ret                     ;

;Text-----------------------------------------------------------
; Installation messages:
; These messages (some of them anyway) are displayed when the
; driver is started. They do not stay resident when the init
; phase of the driver is done.
;---------------------------------------------------------------
;
; This is the normal messages you should always get:
;
InitStr     db 'PoFoCF disk driver v0.02 (02-02-2000)',10,13
            db 'by: Peter Faasse',10,13,'$'
;
; If /D is specified in config.sys you also get:
;
DebugStr    db 'Debug modus active',10,13
            db 'Driver code segment: ','$'
;
; I like to tell what disk I found at the other end of the bus.
; Real ATA-3 disks will give a nice string of what disk is
; attached
;
IdeNameStr  db 'Disk  : ','$'
IdeNoName   db 'Not specified','$'
DrvStr      db 'Drives: d:','$'
DrvStr1     db ' .. ','$'
DrvStr2     db ':','$'
NewLn       db 10,13,'$'
;
;
; some other barking
;
NoFsStr     db 'File-system check skipped',10,13,'$'
FsDnStr     db 'File-system check done',10,13,'$'
;
; Some error messages:
;
; Do I have to explain what went wrong?
;
CmdErrorStr db 'Illegal command line option!',10,13,'$'
NoPoFoStr   db 'Not a Portfolio!',10,13,'$'
SetParErr   db 'Disk Parameter set error',10,13,'$'
NoCfFound   db 'CompactFlash not found',10,13,'$'
NoIdentStr  db 'CompactFlash Ident Failed',10,13,'$'
;
; non-fatal error messages:
;
FsErrStr    db 'No valid filesystem: ','$'
;
;End-------------------------------------------------
;
; some MASM-satisfiers
;
code    ends            ; end of code segment
                        ;
            END
